/*
 * Copyright (c) 2001, Adam Dunkels.
 * SPI2CF version (c) 2006, Till Harbaum.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Adam Dunkels.
 * 4. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * This file is part of the uIP TCP/IP stack.
 */

#include "uip.h"

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <util/delay.h>

#include "uip_arp.h"
#include "spi.h"
#include "cf.h"
#include "prism2.h"

#include "timer.h"
#include "uart.h"
#include "config.h"

#define BUF ((struct uip_eth_hdr *)&uip_buf[0])

#ifndef NULL
#define NULL (void *)0
#endif /* NULL */

/*---------------------------------------------------------------------------*/
int
main(void)
{
  int i;
  struct timer periodic_timer, arp_timer;

  /* enable watchdog */
  wdt_enable(WDTO_2S);

  DDRB |= _BV(2);   // AVR led (green) is output, but still off

  uart_init();

  printf_P(PSTR("\n\nUIP-1.0 - AVR\n"));

  // init everything necessary to drive the prism card
  clock_init();
  spi_init();
  cf_get_id();

  // check if there's a card inserted at all
  if(!cf_get_card_detect())
    config_console_fatal(PSTR("No card inserted"), 4);

  cf_reset();

  /* enabling the config console also makes sure there's a valid */
  /* config stored in eeprom */
  config_init();

  prism2_enable();
  prism2_init();

  timer_set(&periodic_timer, CLOCK_SECOND / 2);
  timer_set(&arp_timer, CLOCK_SECOND * 10);
  
  uip_init();

  /* get config from eeprom */
  if(eeprom_read_byte(&config.flags) & CONFIG_FLAG_DHCP) {
    printf_P(PSTR("Starting dhcp client\n"));
    dhcpc_init(&uip_ethaddr.addr, 6);
  } else {
    uip_ipaddr_t ipaddr;

    printf_P(PSTR("Using static configuration\n"));

    eeprom_read_block(ipaddr, config.addr, 4);
    printf_P(PSTR("Setting IP address %u.%u.%u.%u\n"),
	 uip_ipaddr1(ipaddr), uip_ipaddr2(ipaddr),
	 uip_ipaddr3(ipaddr), uip_ipaddr4(ipaddr));
    uip_sethostaddr(ipaddr);

    eeprom_read_block(ipaddr, config.mask, 4);
    printf_P(PSTR("Setting netmask %u.%u.%u.%u\n"),
	 uip_ipaddr1(ipaddr), uip_ipaddr2(ipaddr),
	 uip_ipaddr3(ipaddr), uip_ipaddr4(ipaddr));
    uip_setnetmask(ipaddr);

    eeprom_read_block(ipaddr, config.gate, 4);
    printf_P(PSTR("Setting default gateway %u.%u.%u.%u\n"),
	 uip_ipaddr1(ipaddr), uip_ipaddr2(ipaddr),
	 uip_ipaddr3(ipaddr), uip_ipaddr4(ipaddr));
    uip_setdraddr(ipaddr);

//    xshow_config();
  }

  httpd_init();
  
  PORTB |= _BV(2);   /* system booted: switch green AVR led on */

  while(1) {
    wdt_reset();     /* calm watch dog */

    /* frequently handle console input */
    config_console_handle();

    uip_len = prism2_poll(uip_buf);
    if(uip_len > 0) {
      if(BUF->type == htons(UIP_ETHTYPE_IP)) {
	uip_arp_ipin();
	uip_input();
	/* If the above function invocation resulted in data that
	   should be sent out on the network, the global variable
	   uip_len is set to a value > 0. */
	if(uip_len > 0) {
	  uip_arp_out();
	  prism2_send(uip_buf, uip_len);
	}
      } else if(BUF->type == htons(UIP_ETHTYPE_ARP)) {
	uip_arp_arpin();
	/* If the above function invocation resulted in data that
	   should be sent out on the network, the global variable
	   uip_len is set to a value > 0. */
	if(uip_len > 0) {
	  prism2_send(uip_buf, uip_len);
	}
      }

    } else if(timer_expired(&periodic_timer)) {
      timer_reset(&periodic_timer);
      for(i = 0; i < UIP_CONNS; i++) {
	uip_periodic(i);
	/* If the above function invocation resulted in data that
	   should be sent out on the network, the global variable
	   uip_len is set to a value > 0. */
	if(uip_len > 0) {
	  uip_arp_out();
	  prism2_send(uip_buf, uip_len);
	}
      }

#if UIP_UDP
      for(i = 0; i < UIP_UDP_CONNS; i++) {
	uip_udp_periodic(i);
	/* If the above function invocation resulted in data that
	   should be sent out on the network, the global variable
	   uip_len is set to a value > 0. */
	if(uip_len > 0) {
	  uip_arp_out();
	  prism2_send(uip_buf, uip_len);
	}
      }
#endif /* UIP_UDP */
      
      /* Call the ARP timer function every 10 seconds. */
      if(timer_expired(&arp_timer)) {
	timer_reset(&arp_timer);
	uip_arp_timer();
      }
    }
  }
  return 0;
}

/*---------------------------------------------------------------------------*/
void
uip_log(char *m)
{
  printf_P(PSTR("uIP log message: "));
  printf_P(m);
  printf_P(PSTR("\n"));
}

#ifdef __DHCPC_H__
void
dhcpc_configured(const struct dhcpc_state *s)
{
  uip_sethostaddr(s->ipaddr);
  uip_setnetmask(s->netmask);
  uip_setdraddr(s->default_router);
}
#endif /* __DHCPC_H__ */
/*---------------------------------------------------------------------------*/
